#include "puzzle.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct {
  int x;
  int y;
  int z;
} pos_t;

int busy[3][3][3];

int normalize_points[3];

void normalize(pos_t *p, pos_t *out, int *norm)
{
  int i;
  int dx;
  int dy;
  int dz;
  int cur;

  /* we need to rotate/translate for points norm[0,1,2] to be at their
   * position, but we won't solve the matrix. Instead we do it piece by
   * piece. First translation, then three rotations.
   */

  /* translation to (0,0,0) for point norm[0] */
  cur = norm[0];
  dx = -p[cur].x;
  dy = -p[cur].y;
  dz = -p[cur].z;
  for (i = 0; i < 27; i++) {
    out[i].x = p[i].x + dx;
    out[i].y = p[i].y + dy;
    out[i].z = p[i].z + dz;
  }

  /* rotation centered at (0,0,0) to (0,0,1) for point norm[1] */
  cur = norm[1];
  /* let's not get the matrix, do it coordinate by coordinate */
  if (out[cur].x) {
    /* pi/2 around z: (x,y,z) goes to (-y,x,z) */
    int nx, ny;
    for (i = 0; i < 27; i++) {
      nx = -out[i].y;
      ny = out[i].x;
      out[i].x = nx;
      out[i].y = ny;
    }
  }
  if (out[cur].y) {
    /* pi/2 around x: (x,y,z) goes to (x,z,-y) */
    int ny, nz;
    for (i = 0; i < 27; i++) {
      ny = out[i].z;
      nz = -out[i].y;
      out[i].y = ny;
      out[i].z = nz;
    }
  }
  if (out[cur].z == -1) {
    /* pi around y: (x,y,z) goes to (-x,y,-z) */
    int nx, nz;
    for (i = 0; i < 27; i++) {
      nx = -out[i].x;
      nz = -out[i].z;
      out[i].x = nx;
      out[i].z = nz;
    }
  }

  /* point norm[2] is now at (1,0,1) or (0,1,1) or (-1,0,1) or (0,-1,1) */
  /* find the rotation that puts it at (0,1,1) */
  //x+iy * a+ib = x*a-y*b   x*b+y*a
  //x+iy * a+ib =   0          1
  // x*a-y*b = 0
  // x*b+y*a = 1
  // x*y*a-y*y*b = 0
  // x*x*b+x*y*a = x
  // b*(x*x+y*y) = x
  // b = x / (x*x+y*y)
  // a = b*y/x or a = 1/y
  cur = norm[2];
  int x = out[cur].x;
  int y = out[cur].y;
  int a, b;
  /* we simplify, x is -1 or 1, so x = 1/x, same for y */
  a = y;
  b = x;
  int nx, ny;
  for (i = 0; i < 27; i++) {
    nx = out[i].x * a - out[i].y * b;
    ny = out[i].x * b + out[i].y * a;
    out[i].x = nx;
    out[i].y = ny;
  }

  /* now move things to 0..2 for all dimensions */
  int minx = out[0].x;
  int miny = out[0].y;
  int minz = out[0].z;
  for (i = 1; i < 27; i++) {
    if (out[i].x < minx) minx = out[i].x;
    if (out[i].y < miny) miny = out[i].y;
    if (out[i].z < minz) minz = out[i].z;
  }
  for (i = 0; i < 27; i++) {
    out[i].x -= minx;
    out[i].y -= miny;
    out[i].z -= minz;
  }
}
int *check_constraints(int cube, int x, int y, int z, int *puzzle, pos_t *p)
{
  int c1, c2, c3, c4;
  int delta;
  int delta_x1, delta_x2;
  int delta_y1, delta_y2;
  int delta_z1, delta_z2;
  int aligned;
  int d1, d2;
  int u1, u2, u3;
  int v1, v2, v3;
  int vec1, vec2, vec3;

  while (*puzzle != CUBE && *puzzle != -1) {
    switch (*puzzle) {
    case TOUCH:
      c1 = puzzle[1];
      delta = abs(p[cube].x - p[c1].x)
            + abs(p[cube].y - p[c1].y)
            + abs(p[cube].z - p[c1].z);
printf("touch p[%d] (%d %d %d) p[%d] (%d %d %d) delta %d\n", cube, p[cube].x, p[cube].y, p[cube].z, c1, p[c1].x, p[c1].y, p[c1].z, delta);
      if (delta != 1) return NULL;
      puzzle += 2;
      break;
    case ALIGNED:
    case NOT_ALIGNED:
      c1 = puzzle[1];
      c2 = puzzle[2];
      delta_x1 = p[c2].x - p[c1].x;
      delta_x2 = p[cube].x - p[c2].x;
      delta_y1 = p[c2].y - p[c1].y;
      delta_y2 = p[cube].y - p[c2].y;
      delta_z1 = p[c2].z - p[c1].z;
      delta_z2 = p[cube].z - p[c2].z;
      aligned = delta_x1 == delta_x2
              && delta_y1 == delta_y2
              && delta_z1 == delta_z2;
printf("%s cube %d c1c2 %d %d aligned %d\n", *puzzle == ALIGNED ? "ALIGNED" : "NOT_ALIGNED", cube, c1, c2, aligned);
      if (*puzzle == ALIGNED && !aligned) return NULL;
      if (*puzzle == NOT_ALIGNED && aligned) return NULL;
      puzzle += 3;
      break;
    case ORDER:
      c1 = puzzle[1];
      c2 = puzzle[2];
      d1 = p[c1].x * 9 + p[c1].y * 3 + p[c1].z;
      d2 = p[c2].x * 9 + p[c2].y * 3 + p[c2].z;
printf("order p[%d] (%d %d %d) p[%d] (%d %d %d) d1 %d d2 %d\n", c1, p[c1].x, p[c1].y, p[c1].z, c2, p[c2].x, p[c2].y, p[c2].z, d1, d2);
      if (d2 >= d1) return NULL;
      puzzle += 3;
      break;
    case LEFT:
      c1 = puzzle[1];
      c2 = puzzle[2];
      c3 = puzzle[3];
      c4 = puzzle[4];
      /* compute vectorial product */
      u1 = p[c2].x - p[c1].x; u2 = p[c2].y - p[c1].y; u3 = p[c2].z - p[c1].z;
      v1 = p[c3].x - p[c1].x; v2 = p[c3].y - p[c1].y; v3 = p[c3].z - p[c1].z;
      vec1 = u2 * v3 - u3 * v2;
      vec2 = u3 * v1 - u1 * v3;
      vec3 = u1 * v2 - u2 * v1;
      /* putting vec at c4 its end should be at cube */
      if (   p[c4].x + vec1 != p[cube].x
          || p[c4].y + vec2 != p[cube].y
          || p[c4].z + vec3 != p[cube].z)
        return NULL;
      puzzle += 5;
      break;
    case VEC:
      c1 = puzzle[1];
      c2 = puzzle[2];
      c3 = puzzle[3];
      vec1 = p[c2].x - p[c1].x;
      vec2 = p[c2].y - p[c1].y;
      vec3 = p[c2].z - p[c1].z;
printf("vec p[%d] (%d %d %d) p[%d] (%d %d %d) p[%d] (%d %d %d) vec(%d %d %d) cube [%d] (%d %d %d) c3+vec (%d %d %d)\n",
       c1, p[c1].x, p[c1].y, p[c1].z, c2, p[c2].x, p[c2].y, p[c2].z, c3, p[c3].x, p[c3].y, p[c3].z, vec1, vec2, vec3, cube, p[cube].x, p[cube].y, p[cube].z,
       p[c3].x + vec1, p[c3].y + vec2, p[c3].z + vec3);
      /* putting vec at c3 its end should be at cube */
      if (   p[c3].x + vec1 != p[cube].x
          || p[c3].y + vec2 != p[cube].y
          || p[c3].z + vec3 != p[cube].z)
        return NULL;
      puzzle += 4;
      break;
    }
  }
  return puzzle;
}

void solve(int cube, pos_t *p, int *puzzle)
{
  int x, y, z;
  int *pz;
  if (cube == 27) {
    int i;
    pos_t p2[27];
    printf("SOLUTION");
    for (i = 0; i < 27; i++) printf(" (%d %d %d %d)", i, p[i].x, p[i].y, p[i].z);
    printf("\n");
    normalize(p, p2, normalize_points);
    printf("solution");
    for (i = 0; i < 27; i++) printf(" (%d %d %d %d)", i, p2[i].x, p2[i].y, p2[i].z);
    printf("\n");
    return;
  }
  if (puzzle[0] != CUBE) abort();
  if (puzzle[1] != cube) abort();
  puzzle += 2;
  for (x = 0; x < 3; x++) for (y = 0; y < 3; y++) for (z = 0; z < 3; z++) {
    if (busy[x][y][z] != -1) continue;
    p[cube].x = x;
    p[cube].y = y;
    p[cube].z = z;
    pz = check_constraints(cube, x, y, z, puzzle, p);
    if (pz == NULL) continue;
    busy[x][y][z] = cube;
    solve(cube + 1, p, pz);
    busy[x][y][z] = -1;
  }
}

void go(int *puzzle)
{
  pos_t p[27];
  int x, y, z;
  for (x = 0; x < 3; x++) for (y = 0; y < 3; y++) for (z = 0; z < 3; z++)
    busy[x][y][z] = -1;

  if (puzzle[0] != NORMALIZE) abort();
  normalize_points[0] = puzzle[1];
  normalize_points[1] = puzzle[2];
  normalize_points[2] = puzzle[3];

  solve(0, p, puzzle + 4);
}

int main(void)
{
  go(puzzle_3);
  return 0;
}
